@@ -11,31 +11,66 @@ local log = require "util.logger".init("stanzarouter")
 local hosts = _G.prosody.hosts;
 local tostring = tostring;
 local st = require "util.stanza";
-local send_s2s = require "core.s2smanager".send_to_host;
-local modules_handle_stanza = require "core.modulemanager".handle_stanza;
-local component_handle_stanza = require "core.componentmanager".handle_stanza;
 local jid_split = require "util.jid".split;
 local jid_prepped_split = require "util.jid".prepped_split;
 
 local full_sessions = _G.prosody.full_sessions;
 local bare_sessions = _G.prosody.bare_sessions;
 
+local core_post_stanza, core_process_stanza, core_route_stanza;
+
+function deprecated_warning(f)
+	_G[f] = function(...)
+		log("warn", "Using the global %s() is deprecated, use module:send() or prosody.%s(). %s", f, f, debug.traceback());
+		return prosody[f](...);
+	end
+end
+deprecated_warning"core_post_stanza";
+deprecated_warning"core_process_stanza";
+deprecated_warning"core_route_stanza";
+
+local valid_stanzas = { message = true, presence = true, iq = true };
+local function handle_unhandled_stanza(host, origin, stanza)
+	local name, xmlns, origin_type = stanza.name, stanza.attr.xmlns or "jabber:client", origin.type;
+	if xmlns == "jabber:client" and valid_stanzas[name] then
+		-- A normal stanza
+		local st_type = stanza.attr.type;
+		if st_type == "error" or (name == "iq" and st_type == "result") then
+			log("debug", "Discarding %s from %s of type: %s", name, origin_type, st_type or '<nil>');
+			return;
+		end
+		if name == "iq" and (st_type == "get" or st_type == "set") and stanza.tags[1] then
+			xmlns = stanza.tags[1].attr.xmlns or "jabber:client";
+		end
+		log("debug", "Unhandled %s stanza: %s; xmlns=%s", origin_type, name, xmlns);
+		if origin.send then
+			origin.send(st.error_reply(stanza, "cancel", "service-unavailable"));
+		end
+	elseif not((name == "features" or name == "error") and xmlns == "http://etherx.jabber.org/streams") then -- FIXME remove check once we handle S2S features
+		log("warn", "Unhandled %s stream element or stanza: %s; xmlns=%s: %s", origin_type, name, xmlns, tostring(stanza)); -- we didn't handle it
+		origin:close("unsupported-stanza-type");
+	end
+end
+
+local iq_types = { set=true, get=true, result=true, error=true };
 function core_process_stanza(origin, stanza)
 	(origin.log or log)("debug", "Received[%s]: %s", origin.type, stanza:top_tag())
 
-	-- TODO verify validity of stanza (as well as JID validity)
-	if stanza.attr.type == "error" and #stanza.tags == 0 then return; end -- TODO invalid stanza, log
-	if stanza.name == "iq" then
-		if not stanza.attr.id then stanza.attr.id = ""; end -- COMPAT Jabiru doesn't send the id attribute on roster requests
-		if (stanza.attr.type == "set" or stanza.attr.type == "get") and (#stanza.tags ~= 1) then
-			origin.send(st.error_reply(stanza, "modify", "bad-request"));
-			return;
-		end
-	end
-
 	if origin.type == "c2s" and not stanza.attr.xmlns then
+		local name, st_type = stanza.name, stanza.attr.type;
+		if st_type == "error" and #stanza.tags == 0 then
+			return handle_unhandled_stanza(origin.host, origin, stanza);
+		end
+		if name == "iq" then
+			if not stanza.attr.id then stanza.attr.id = ""; end -- COMPAT Jabiru doesn't send the id attribute on roster requests
+			if not iq_types[st_type] or (st_type ~= "result" and #stanza.tags ~= 1) then
+				origin.send(st.error_reply(stanza, "modify", "bad-request", "Invalid IQ type or incorrect number of children"));
+				return;
+			end
+		end
+
 		if not origin.full_jid
-			and not(stanza.name == "iq" and stanza.attr.type == "set" and stanza.tags[1] and stanza.tags[1].name == "bind"
+			and not(name == "iq" and st_type == "set" and stanza.tags[1] and stanza.tags[1].name == "bind"
 					and stanza.tags[1].attr.xmlns == "urn:ietf:params:xml:ns:xmpp-bind") then
 			-- authenticated client isn't bound and current stanza is not a bind request
 			if stanza.attr.type ~= "result" and stanza.attr.type ~= "error" then
@@ -84,17 +119,17 @@ function core_process_stanza(origin, stanza)
 		stanza.attr.from = from;
 	end
 
-	--[[if to and not(hosts[to]) and not(hosts[to_bare]) and (hosts[host] and hosts[host].type ~= "local") then -- not for us?
-		log("warn", "stanza recieved for a non-local server");
-		return; -- FIXME what should we do here?
-	end]] -- FIXME
-
 	if (origin.type == "s2sin" or origin.type == "c2s" or origin.type == "component") and xmlns == nil then
 		if origin.type == "s2sin" and not origin.dummy then
 			local host_status = origin.hosts[from_host];
 			if not host_status or not host_status.authed then -- remote server trying to impersonate some other server?
 				log("warn", "Received a stanza claiming to be from %s, over a stream authed for %s!", from_host, origin.from_host);
-				return; -- FIXME what should we do here? does this work with subdomains?
+				origin:close("not-authorized");
+				return;
+			elseif not hosts[host] then
+				log("warn", "Remote server %s sent us a stanza for %s, closing stream", origin.from_host, host);
+				origin:close("host-unknown");
+				return;
 			end
 		end
 		core_post_stanza(origin, stanza, origin.full_jid);
@@ -114,7 +149,7 @@ function core_process_stanza(origin, stanza)
 			if h.events.fire_event(event, {origin = origin, stanza = stanza}) then return; end
 		end
 		if host and not hosts[host] then host = nil; end -- COMPAT: workaround for a Pidgin bug which sets 'to' to the SRV result
-		modules_handle_stanza(host or origin.host or origin.to_host, origin, stanza);
+		handle_unhandled_stanza(host or origin.host or origin.to_host, origin, stanza);
 	end
 end
 
@@ -123,7 +158,7 @@ function core_post_stanza(origin, stanza, preevents)
 	local node, host, resource = jid_split(to);
 	local to_bare = node and (node.."@"..host) or host; -- bare JID
 
-	local to_type;
+	local to_type, to_self;
 	if node then
 		if resource then
 			to_type = '/full';
@@ -131,6 +166,7 @@ function core_post_stanza(origin, stanza, preevents)
 			to_type = '/bare';
 			if node == origin.username and host == origin.host then
 				stanza.attr.to = nil;
+				to_self = true;
 			end
 		end
 	else
@@ -138,6 +174,7 @@ function core_post_stanza(origin, stanza, preevents)
 			to_type = '/host';
 		else
 			to_type = '/bare';
+			to_self = true;
 		end
 	end
 
@@ -148,12 +185,8 @@ function core_post_stanza(origin, stanza, preevents)
 	local h = hosts[to_bare] or hosts[host or origin.host];
 	if h then
 		if h.events.fire_event(stanza.name..to_type, event_data) then return; end -- do processing
-
-		if h.type == "component" then
-			component_handle_stanza(origin, stanza);
-			return;
-		end
-		modules_handle_stanza(h.host, origin, stanza);
+		if to_self and h.events.fire_event(stanza.name..'/self', event_data) then return; end -- do processing
+		handle_unhandled_stanza(h.host, origin, stanza);
 	else
 		core_route_stanza(origin, stanza);
 	end
@@ -170,26 +203,24 @@ function core_route_stanza(origin, stanza)
 	if hosts[host] then
 		-- old stanza routing code removed
 		core_post_stanza(origin, stanza);
-	elseif origin.type == "c2s" then
-		-- Remote host
-		if not hosts[from_host] then
-			log("error", "No hosts[from_host] (please report): %s", tostring(stanza));
-		end
-		if (not hosts[from_host]) or (not hosts[from_host].disallow_s2s) then
-			local xmlns = stanza.attr.xmlns;
-			--stanza.attr.xmlns = "jabber:server";
-			stanza.attr.xmlns = nil;
-			log("debug", "sending s2s stanza: %s", tostring(stanza));
-			send_s2s(origin.host, host, stanza); -- TODO handle remote routing errors
-			stanza.attr.xmlns = xmlns; -- reset
-		else
-			core_route_stanza(hosts[from_host], st.error_reply(stanza, "cancel", "not-allowed", "Communication with remote servers is not allowed"));
-		end
-	elseif origin.type == "component" or origin.type == "local" then
-		-- Route via s2s for components and modules
-		log("debug", "Routing outgoing stanza for %s to %s", from_host, host);
-		send_s2s(from_host, host, stanza);
 	else
-		log("warn", "received %s stanza from unhandled connection type: %s", tostring(stanza.name), tostring(origin.type));
+		log("debug", "Routing to remote...");
+		local host_session = hosts[from_host];
+		if not host_session then
+			log("error", "No hosts[from_host] (please report): %s", tostring(stanza));
+		else
+			local xmlns = stanza.attr.xmlns;
+			stanza.attr.xmlns = nil;
+			local routed = host_session.events.fire_event("route/remote", { origin = origin, stanza = stanza, from_host = from_host, to_host = host });
+			stanza.attr.xmlns = xmlns; -- reset
+			if not routed then
+				log("debug", "... no, just kidding.");
+				if stanza.attr.type == "error" or (stanza.name == "iq" and stanza.attr.type == "result") then return; end
+				core_route_stanza(host_session, st.error_reply(stanza, "cancel", "not-allowed", "Communication with remote domains is not enabled"));
+			end
+		end
 	end
 end
+prosody.core_process_stanza = core_process_stanza;
+prosody.core_post_stanza = core_post_stanza;
+prosody.core_route_stanza = core_route_stanza;
