@@ -11,11 +11,11 @@ local welcome_text = module:get_option("welcome_message") or "Hello $username, w
 
 local st = require "util.stanza";
 
-module:hook("user-registered", 
+module:hook("user-registered",
 	function (user)
-		local welcome_stanza = 
+		local welcome_stanza =
 			st.message({ to = user.username.."@"..user.host, from = host })
 				:tag("body"):text(welcome_text:gsub("$(%w+)", user));
-		core_route_stanza(hosts[host], welcome_stanza);
+		module:send(welcome_stanza);
 		module:log("debug", "Welcomed user %s@%s", user.username, user.host);
 	end);
